import json

import aiojobs
import aio_pika
from aio_pika.patterns import RPC

from typing import Callable, Awaitable


class JsonRPC(RPC):
    SERIALIZER = json
    CONTENT_TYPE = 'application/json'

    def serialize(self, data) -> bytes:
        return self.SERIALIZER.dumps(data, ensure_ascii=False, default=repr).encode()

    def serialize_exception(self, exception: Exception) -> bytes:
        return self.serialize({
            "error": {
                "type": exception.__class__.__name__,
                "message": repr(exception),
                "args": exception.args,
            }
        })



class Broker(object):
    """
    On RPC errors return:
    ```python
    return self.serialize({
            "error": {
                "type": exception.__class__.__name__,
                "message": repr(exception),
                "args": exception.args,
            }
        })
    ```
    """
    def __init__(self) -> None:
        self._connection = None
        self.scheduler = None
        self._rpc_functions = dict()
        self._rpc_call_channel = None
        self._rpc_client = None
        self._rpc_server_channel = None
        self._rpc_server = None
        self._event_consumer_channel = None
        self._event_publisher_channel = None
        self._direct_consumer_channel = None
        self._direct_publisher_channel = None
        self._exchanges = dict()

    def add_rpc_func(self, name, func):
        self._rpc_functions[name] = func
        
    async def register_rpc_functions(self):
        for name, func in self._rpc_functions.items():
            await self.register_rpc_function(name, func)
            
    async def init_event_decorators(self):
        self.scheduler = await aiojobs.create_scheduler()

    async def close(self):
        await self._connection.close()
        
    def add_close_callback(self, *args, **kwargs):
        self._connection.add_close_callback(*args, **kwargs)
    
    def add_reconnect_callback(self, *args, **kwargs):
        self._connection.add_reconnect_callback(*args, **kwargs)

    @property
    def connection(self) -> aio_pika.connection.Connection:
        return self._connection

    async def connect(self, *args, **kwargs) -> None:
        self._connection = await aio_pika.connect_robust(*args, **kwargs)
    
    async def get_rpc_client(self):
        """
        Get or create RPC client
        """
        if self._rpc_call_channel is None:
            self._rpc_call_channel = await self._connection.channel()
            if self._rpc_client is None:
                self._rpc_client = await JsonRPC.create(self._rpc_call_channel)
                return self._rpc_client
            await self._rpc_client.initialize()
        elif self._rpc_client is None:
            self._rpc_client = await JsonRPC.create(self._rpc_call_channel)
            return self._rpc_client
        return self._rpc_client
    
    def is_registered(self, func):
        return self._rpc_server and func in self._rpc_server.consumer_tags
    
    # async def register_rpc_function(self, name: str, function: Callable[[...], Awaitable[...]]) -> None:
    async def register_rpc_function(self, name: str, function) -> None:
        if self._rpc_server is None:
            if self._rpc_server_channel is None:
                self._rpc_server_channel = await self._connection.channel()
            self._rpc_server = await JsonRPC.create(self._rpc_server_channel)
        await self._rpc_server.register(name, function, auto_delete=False)
    
    async def publish_event(self,
                            exchange_name: str,
                            routing_key: str,
                            data: dict,
                            durable: bool = True,
                            content_type: str = 'application/json',
                            content_encoding: str = 'utf-8') -> None:
        if self._event_publisher_channel is None:
            self._event_publisher_channel = await self._connection.channel()
        if exchange_name in self._exchanges:
            exchange = self._exchanges[exchange_name]
        else:
            exchange = await self._event_publisher_channel.declare_exchange(
                name=exchange_name,
                type=aio_pika.ExchangeType.TOPIC,
                durable=durable,
                auto_delete=False,
                internal=False,
                passive=False
            )
            self._exchanges[exchange_name] = exchange
        message = aio_pika.Message(
            body=json.dumps(data).encode(),
            content_type=content_type,
            content_encoding=content_encoding,
            delivery_mode=aio_pika.DeliveryMode.PERSISTENT,
        )
        await exchange.publish(
            message,
            routing_key=routing_key
        )
    
    async def register_event_consumer(self,
                                      exchange_name: str,
                                      queue_name: str,
                                      routing_key: str,
                                      callback,
                                      durable: bool = True,
                                      no_ack: bool = False) -> None:
        if self._event_consumer_channel is None:
            self._event_consumer_channel = await self._connection.channel()
        exchange = await self._event_consumer_channel.declare_exchange(
            name=exchange_name,
            type=aio_pika.ExchangeType.TOPIC,
            durable=durable,
            auto_delete=False,
            internal=False,
            passive=False
        )
        queue = await self._event_consumer_channel.declare_queue(
            name=queue_name,
            durable=durable,
            exclusive=False,
            passive=False,
            auto_delete=False
        )
        await queue.bind(exchange, routing_key=routing_key)
        await queue.consume(
            callback=callback,
            no_ack=no_ack,
            exclusive=False,
        )
        
    async def publish_direct(self,
                            exchange_name: str,
                            direct: str,
                            data: dict,
                            durable: bool = True,
                            content_type: str = 'application/json',
                            content_encoding: str = 'utf-8') -> None:
        if self._direct_publisher_channel is None:
            self._direct_publisher_channel = await self._connection.channel()
        if exchange_name in self._exchanges:
            exchange = self._exchanges[exchange_name]
        else:
            exchange = await self._direct_publisher_channel.declare_exchange(
                name=exchange_name,
                type=aio_pika.ExchangeType.DIRECT,
                durable=durable,
                auto_delete=False,
                internal=False,
                passive=False
            )
            self._exchanges[exchange_name] = exchange
        message = aio_pika.Message(
            body=json.dumps(data).encode(),
            content_type=content_type,
            content_encoding=content_encoding,
            delivery_mode=aio_pika.DeliveryMode.NOT_PERSISTENT,
        )
        await exchange.publish(
            message,
            routing_key=direct
        )
    
    async def register_direct_consumer(self,
                                      exchange_name: str,
                                      queue_name: str,
                                      direct: str,
                                      callback,
                                      durable: bool = True,
                                      no_ack: bool = True) -> None:
        if self._direct_consumer_channel is None:
            self._direct_consumer_channel = await self._connection.channel()
        exchange = await self._direct_consumer_channel.declare_exchange(
            name=exchange_name,
            type=aio_pika.ExchangeType.DIRECT,
            durable=durable,
            auto_delete=False,
            internal=False,
            passive=False
        )
        queue = await self._event_consumer_channel.declare_queue(
            name=queue_name,
            durable=durable,
            exclusive=False,
            passive=False,
            auto_delete=False
        )
        await queue.bind(exchange, routing_key=direct)
        await queue.consume(
            callback=callback,
            no_ack=no_ack,
            exclusive=False,
        )



import functools
import inspect

import aiojobs


def rpc_method(broker, name=None):
    def rpc_method_decorator(func, name=name):
        if name is None:
            name = func.__name__
            # name = '_'.join(func.__qualname__.split('.')[2:])
        broker.add_rpc_func(name, func)
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return  wrapper
    return rpc_method_decorator


def event(broker, exchange, routing_key):
    def event_decorator(func, exchange=exchange, routing_key=routing_key):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            event_context = dict()
            
            argnames, _, _, defaults = inspect.getargspec(func) 
            
            if defaults:
                event_context.update(dict(zip(reversed(argnames), reversed(defaults))))
            if args:
                event_context.update(dict(zip(argnames, args)))
            event_context.update(kwargs)
            if 'self' in event_context:
                del event_context['self']
            if 'event_context' in argnames:
                del event_context['event_context']
                kwargs['event_context'] = event_context
            result = await func(*args, **kwargs)
            
            await broker.scheduler.spawn(broker.publish_event(exchange, routing_key, event_context))
            
            return result
        return  wrapper
    return event_decorator


from setuptools import setup, find_packages

setup(
  name = 'AioBrokerTools',
  packages=find_packages(),
  version = '0.1.1',
  license='MIT',
  description = 'Async rabbitmq microservices framework',
  author = 'Maksim Vorontsov',
  author_email = 'social.maksim.vrs@gmail.com',
  url = 'https://gitlab.com/maksimvrs/aiobroker',
  download_url = 'https://gitlab.com/maksimvrs/aiobroker/-/archive/master/aiobroker-master.tar.gz',
  keywords = ['async', 'broker', 'rabbitmq', 'framework', 'microservice'],
  install_requires=[
        'aio-pika',
        'aiojobs',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ],
)